import collections
import copy
import dataclasses
import enum
import html.parser
import os
import pathlib
import re
import threading
import typing
from . import certificates
from . import progress
from . import utilities

certificates_bundle = certificates.bundle()
if certificates_bundle is not None:
    os.environ["REQUESTS_CA_BUNDLE"] = certificates_bundle
import requests


@dataclasses.dataclass
class Resource:
    local_path: pathlib.Path
    remote_path: pathlib.PurePosixPath

    @classmethod
    def from_string(
        cls, local_path: typing.Union[str, pathlib.Path], remote_path: typing.Union[str, pathlib.PurePosixPath]
    ) -> "Resource":
        if isinstance(local_path, str):
            local_path = pathlib.Path(local_path)
        if isinstance(remote_path, str):
            remote_path = pathlib.PurePosixPath(remote_path)
        return cls(local_path=local_path, remote_path=remote_path)

    def as_lzip(self) -> "Resource":
        return Resource(
            local_path=pathlib.Path(str(self.local_path) + ".lz"),
            remote_path=pathlib.PurePosixPath(str(self.remote_path) + ".lz"),
        )

    def download_path(self) -> pathlib.Path:
        return pathlib.Path(str(self.local_path) + ".download")

    def download_todo(self, force: bool) -> tuple[bool, int]:
        if force:
            return (True, 0)
        if self.local_path.is_file():
            return (False, self.local_path.stat().st_size)
        if self.download_path().is_file():
            return (True, self.download_path().stat().st_size)
        return (True, 0)


@dataclasses.dataclass
class Task(Resource):
    class State(enum.Enum):
        UNKNOWN = 0
        TODO = 1
        DOING = 2
        DONE = 3

    state: State = State.UNKNOWN
    to_download_size: int = 0
    downloaded_size: int = 0


class Workload:
    def __init__(self):
        self.status = progress.Status()
        self.status_lock = threading.Lock()
        self.tasks = collections.deque()
        self.done_tasks = collections.deque()
        self.progress_queue: typing.Optional[collections.deque] = None

    def update_status(self, **kwargs):
        status_update = progress.StatusUpdate(**kwargs)
        with self.status_lock:
            self.status.apply_update(status_update=status_update)
        if self.progress_queue is not None:
            self.progress_queue.append(status_update)

    def add_todo_task(self, resource: Resource, size: int) -> None:
        self.tasks.append(
            Task(
                local_path=resource.local_path,
                remote_path=resource.remote_path,
                state=Task.State.TODO,
                to_download_size=size,
                downloaded_size=0,
            )
        )
        self.update_status(
            todo_file_count_delta=1,
            todo_size_delta=size,
        )

    def add_doing_task(self, resource: Resource, to_download_size: int, downloaded_size: int) -> None:
        self.tasks.append(
            Task(
                local_path=resource.local_path,
                remote_path=resource.remote_path,
                state=Task.State.DOING,
                to_download_size=to_download_size,
                downloaded_size=downloaded_size,
            )
        )
        self.update_status(
            todo_file_count_delta=1,
            todo_size_delta=to_download_size,
            done_size_delta=downloaded_size,
        )

    def add_done_task(self, resource: Resource, size: int) -> None:
        self.done_tasks.append(
            Task(
                local_path=resource.local_path,
                remote_path=resource.remote_path,
                state=Task.State.DONE,
                to_download_size=0,
                downloaded_size=size,
            )
        )
        self.update_status(done_file_count_delta=1, done_size_delta=size)

    def resize_task(self, task: Task, actual_to_download_size: int, actual_downloaded_size: int) -> None:
        if task.to_download_size == actual_to_download_size:
            to_download_delta = 0
        else:
            to_download_delta = actual_to_download_size - task.to_download_size
            task.to_download_size = actual_to_download_size
        if task.downloaded_size == actual_downloaded_size:
            downloaded_delta = 0
        else:
            downloaded_delta = actual_downloaded_size - task.downloaded_size
            task.downloaded_size = actual_downloaded_size
        if to_download_delta != 0 or downloaded_delta != 0:
            self.update_status(todo_size_delta=to_download_delta, done_size_delta=downloaded_delta)

    def report_downloaded(self, task: Task, size: int) -> None:
        task.to_download_size = max(0, task.to_download_size - size)
        task.downloaded_size += size
        self.update_status(todo_size_delta=-size, done_size_delta=size)

    def report_done(self, task: Task) -> None:
        if task.to_download_size == 0:
            delta = 0
        else:
            delta = -task.to_download_size
        task.to_download_size += delta
        task.state = Task.State.DONE
        self.done_tasks.append(task)
        self.update_status(
            todo_size_delta=delta,
            todo_file_count_delta=-1,
            done_file_count_delta=1,
        )


class Server:
    def __init__(self, url: str, timeout: float):
        self._url: str = url
        self._timeout: float = timeout

    def __repr__(self):
        return f"{self.__class__.__name__}('{self._url}')"

    def set_timeout(self, timeout: float) -> None:
        self._timeout = timeout

    def path_to_url(self, remote_path: pathlib.PurePosixPath) -> str:
        if remote_path.parts == ():
            return self._url
        return f"{self._url}{remote_path}"

    def join_url(self, url: str, trailing_slash: bool) -> str:
        if trailing_slash and not url.endswith("/"):
            return f"{self._url}{url}/"
        return f"{self._url}{url}"

    def clone_with_url(self, url: str) -> "Server":
        clone = copy.copy(self)
        clone._url = url
        return clone

    def clear_cache(self) -> None:
        pass

    def session(self) -> requests.Session:
        return requests.Session()

    def accept_ranges_and_size(
        self, session: requests.Session, resource: Resource
    ) -> tuple[bool, typing.Optional[int]]:
        response = session.head(
            self.path_to_url(resource.remote_path),
            allow_redirects=True,
            timeout=self._timeout,
        )
        if response.status_code == 404:
            return (False, None)
        response.raise_for_status()
        return (
            "Accept-Ranges" in response.headers and response.headers["Accept-Ranges"] == "bytes",
            int(response.headers["Content-Length"]),
        )

    def resource_pick_failed(self, resources: collections.abc.Sequence[Resource]):
        raise RuntimeError(
            "error 404 (not found) returned by all the candidates ({})".format(
                ", ".join(self.path_to_url(resource.remote_path) for resource in resources)
            )
        )

    def resource_pick_from(
        self, session: requests.Session, resources: collections.abc.Sequence[Resource], force: bool
    ) -> tuple[Resource, int, int]:
        selected_resource = None
        selected_downloaded_size = None
        for resource in resources:
            must_download, downloaded_size = resource.download_todo(force)
            if must_download:
                if downloaded_size > 0:
                    selected_resource = resource
                    selected_downloaded_size = downloaded_size
                    break
            else:
                return (resource, 0, downloaded_size)
        if selected_resource is None:
            for resource in resources:
                _, total_size = self.accept_ranges_and_size(session=session, resource=resource)
                if total_size is not None:
                    return (resource, total_size, 0)
        else:
            accept_ranges, total_size = self.accept_ranges_and_size(session=session, resource=selected_resource)
            if accept_ranges:
                return (selected_resource, max(0, total_size - selected_downloaded_size), selected_downloaded_size)
            return (selected_resource, total_size, 0)
        self.resource_pick_failed(resources=resources)

    def resource_pick(
        self, session: requests.Session, resource: Resource, try_alternatives: bool
    ) -> tuple[Resource, int, int]:
        return self.resource_pick_from(
            session=session,
            resources=(resource, resource.as_lzip()) if try_alternatives else (resource,),
            force=True,
        )

    def resource_size_and_chunks(self, session: requests.Session, resource: Resource) -> typing.Iterator[bytes]:
        response = session.get(
            self.path_to_url(resource.remote_path),
            timeout=self._timeout,
            stream=True,
        )
        response.raise_for_status()
        actual_to_download_size = int(response.headers["Content-Length"])

        def chunks(chunk_size: int) -> typing.Iterator[bytes]:
            for chunk in response.iter_content(chunk_size=chunk_size):
                yield chunk

        return (actual_to_download_size, chunks)

    def workload(
        self,
        resources: collections.abc.Iterable[Resource],
        force: bool,
        try_alternatives: bool,
        workers_count: int,
    ) -> Workload:
        resources = collections.deque(resources)
        workload = Workload()

        def worker_target():
            with self.session() as local_session:
                try:
                    while True:
                        resource = resources.popleft()
                        resource, to_download_size, downloaded_size = self.resource_pick_from(
                            session=local_session,
                            resources=(resource, resource.as_lzip()) if try_alternatives else (resource,),
                            force=force,
                        )
                        if to_download_size > 0:
                            if downloaded_size > 0:
                                workload.add_doing_task(
                                    resource=resource,
                                    to_download_size=to_download_size,
                                    downloaded_size=downloaded_size,
                                )
                            else:
                                workload.add_todo_task(resource=resource, size=to_download_size)
                        elif downloaded_size > 0:
                            workload.add_done_task(resource=resource, size=downloaded_size)
                        else:
                            workload.add_todo_task(resource=resource, size=0)
                except IndexError:
                    pass

        if min(workers_count, len(resources)) < 2:
            worker_target()
        else:
            workers = []
            for _ in range(0, min(workers_count, len(resources))):
                worker = threading.Thread(target=worker_target)
                worker.daemon = True
                worker.start()
                workers.append(worker)
            for worker in workers:
                worker.join()
            del workers
        del resources
        del worker_target
        return workload

    def consume(
        self,
        workload: Workload,
        workers_count: int,
    ) -> None:
        def worker_target():
            with self.session() as local_session:
                try:
                    while True:
                        task = workload.tasks.popleft()
                        assert task.state != Task.State.UNKNOWN
                        assert task.state != Task.State.DONE
                        download_path = task.download_path()
                        if task.state == Task.State.TODO:
                            todo = True
                        elif task.state == Task.State.DOING:
                            actual_downloaded_size = download_path.stat().st_size
                            mode = "ab"
                            response = local_session.get(
                                self.path_to_url(task.remote_path),
                                timeout=self._timeout,
                                stream=True,
                                headers={"Range": f"bytes={actual_downloaded_size}-"},
                            )
                            if response.status_code == 416:
                                todo = True
                            else:
                                todo = False
                                response.raise_for_status()
                                if response.status_code != 206:
                                    raise RuntimeError(
                                        "unexpected status {} for URL {} (expected 206)".format(
                                            response.status, self.path_to_url(task.remote_path)
                                        )
                                    )
                        else:
                            raise Exception(f"unexpected task state {task.state}")
                        if todo:
                            download_path.unlink(missing_ok=True)
                            actual_downloaded_size = 0
                            mode = "wb"
                            response = local_session.get(
                                self.path_to_url(task.remote_path),
                                timeout=self._timeout,
                                stream=True,
                            )
                            response.raise_for_status()
                        actual_to_download_size = int(response.headers["Content-Length"])
                        workload.resize_task(task, actual_to_download_size, actual_downloaded_size)
                        with open(download_path, mode) as doing_file:
                            for chunk in response.iter_content(chunk_size=65536):
                                doing_file.write(chunk)
                                workload.report_downloaded(task, len(chunk))
                        download_path.rename(task.local_path)
                        workload.report_done(task)
                except IndexError:
                    pass

        if min(workers_count, len(workload.tasks)) < 2:
            worker_target()
        else:
            workers = []
            for _ in range(0, min(workers_count, len(workload.tasks))):
                worker = threading.Thread(target=worker_target)
                worker.daemon = True
                worker.start()
                workers.append(worker)
            for worker in workers:
                worker.join()
            del workers
        del worker_target
        del workload

    def download(self, resource: Resource, force: bool, try_alternatives: bool) -> None:
        workload = self.workload(resources=(resource,), force=force, try_alternatives=try_alternatives, workers_count=1)
        self.consume(workload, workers_count=1)


class LocalServer(Server):
    class LocalSession:
        def head(self, url: str, allow_redirects: bool, timeout: float):
            raise FileNotFoundError(url)

        def get(self, url: str, timeout: float, stream: bool, headers: dict[str, str]):
            raise FileNotFoundError(url)

        def __enter__(self):
            return self

        def __exit__(self, *_):
            pass

    def session(self):
        return self.__class__.LocalSession()


class FileServer(Server):
    class Parser:
        def __init__(self):
            raise NotImplementedError("FileServer children must define a Parser subclass")

    def __init__(self, url: str, timeout: float):
        super().__init__(url, timeout)
        self._parent_to_name_to_size: dict[str, int] = {}

    def clone_with_url(self, url: str) -> "Server":
        clone = copy.copy(self)
        clone._url = url
        clone._parent_to_name_to_size = {}
        return clone

    def clear_cache(self) -> None:
        self._parent_to_name_to_size = {}

    def accept_ranges_and_size(
        self, session: requests.Session, resource: Resource
    ) -> tuple[bool, typing.Optional[int]]:
        parent_url = self.path_to_url(resource.remote_path.parent)
        if not resource.remote_path.parent in self._parent_to_name_to_size:
            response = session.get(parent_url, timeout=self._timeout)
            response.raise_for_status()
            parser = self.__class__.Parser()
            parser.feed(response.text)
            parser.close()
            self._parent_to_name_to_size[resource.remote_path.parent] = parser.name_to_size
        if resource.remote_path.name in self._parent_to_name_to_size[resource.remote_path.parent]:
            return (True, self._parent_to_name_to_size[resource.remote_path.parent][resource.remote_path.name])
        return (False, None)

    def workload(
        self,
        resources: collections.abc.Iterable[Resource],
        force: bool,
        try_alternatives: bool,
        workers_count: int,
    ) -> Workload:
        return super().workload(resources=resources, force=force, try_alternatives=try_alternatives, workers_count=1)

    def resource_pick_failed(self, resources: collections.abc.Sequence[Resource]):
        raise RuntimeError(
            "none of the candidates are listed in their parent directory ({})".format(
                ", ".join(
                    f"{resource.remote_path.name} in {self.path_to_url(resource.remote_path.parent)}"
                    for resource in resources
                )
            )
        )


class ApacheServer(FileServer):
    class Parser(html.parser.HTMLParser):
        pattern = re.compile(r"^\s*(\d+(?:\.\d+)?[KMGTP]?)\s*$")

        def __init__(self):
            super().__init__()
            self.name_to_size: dict[str, int] = {}
            self.current_href: typing.Optional[str] = None
            self.state: int = 0

        def handle_starttag(self, tag, attrs):
            if tag == "a":
                try:
                    href = next(value for name, value in attrs if name == "href")
                    self.current_href = href
                    self.state = 1
                except StopIteration:
                    pass
            elif self.state == 1:
                if tag == "td":
                    self.state = 2
            elif self.state == 2:
                if tag == "td":
                    self.state = 3

        def handle_data(self, data):
            if self.state == 1:
                if data == "Parent Directory":
                    self.state = 0
            elif self.state == 3:
                match = self.__class__.pattern.match(data)
                if match is not None:
                    self.name_to_size[self.current_href] = utilities.parse_size(match[1])
                self.state = 0


class NginxServer(FileServer):
    class Parser(html.parser.HTMLParser):
        pattern = re.compile(r"^.*\s+(\d+(?:\.\d+)?[KMGTP]?)\s*$")

        def __init__(self):
            super().__init__()
            self.name_to_size: dict[str, int] = {}
            self.current_href: typing.Optional[str] = None
            self.state: int = 0

        def handle_starttag(self, tag, attrs):
            self.state = 0
            if tag == "a":
                try:
                    href = next(value for name, value in attrs if name == "href")
                    if href != "../":
                        self.current_href = href
                        self.state = 1
                except StopIteration:
                    pass

        def handle_data(self, data):
            if self.state == 1:
                self.state = 2
            elif self.state == 2:
                match = self.__class__.pattern.match(data)
                if match is not None:
                    self.name_to_size[self.current_href] = utilities.parse_size(match[1])
                self.state = 0



import functools
import types
import numpy

def aps_dtype(width, height):
    return [
        ('t', '<u8'),
        ('begin_t', '<u8'),
        ('end_t', '<u8'),
        ('exposure_begin_t', '<u8'),
        ('exposure_end_t', '<u8'),
        ('format', '<u8'),
        ('width', '<u2'),
        ('height', '<u2'),
        ('offset_x', '<i2'),
        ('offset_y', '<i2'),
        ('pixels', (numpy.dtype('<u2'), (width, height))),
    ]

for name, dtype in (
    ('dvs', [('t', '<u8'), ('x', '<u2'), ('y', '<u2'), ('p', '<u1')]),
    ('aps_240_180', aps_dtype(240, 180)),
    ('imu', [
        ('t', '<u8'),
        ('temperature', '<f4'),
        ('accelerometer_x', '<f4'),
        ('accelerometer_y', '<f4'),
        ('accelerometer_z', '<f4'),
        ('gyroscope_x', '<f4'),
        ('gyroscope_y', '<f4'),
        ('gyroscope_z', '<f4'),
        ('magnetometer_x', '<f4'),
        ('magnetometer_y', '<f4'),
        ('magnetometer_z', '<f4'),
    ]),
):
    namespace = types.SimpleNamespace()
    globals()[name] = namespace
    namespace.dtype = numpy.dtype(dtype)
    namespace.zeros = functools.partial(numpy.zeros, dtype=namespace.dtype)
    namespace.fromfile = functools.partial(numpy.fromfile, dtype=namespace.dtype)
    namespace.frombuffer = functools.partial(numpy.frombuffer, dtype=namespace.dtype)



import argparse
import json
import numpy
import pathlib
import shutil
from . import *

dirname = pathlib.Path(__file__).resolve().parent


def check_positive(value):
    value_as_int = int(value)
    if value_as_int <= 0:
        raise argparse.ArgumentTypeError(f"{value} is not a positive integer")
    return value_as_int


def configuration_and_timeout(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--configuration", "-c", default="undr.toml", help="UNDR configuration file path")
    parser.add_argument("--timeout", "-t", default=None, type=float, help="Socket timeout in seconds")


def check_aps_file(file: ApsFile) -> set[str]:
    errors = set()
    if file.width is None:
        errors.add("missing width")
    if file.height is None:
        errors.add("missing height")
    if len(errors) == 0:
        non_monotonic_ts = 0
        size_mismatches = 0
        empty = True
        try:
            previous_t = 0
            for frames in file.packets():
                empty = False
                if len(frames) > 0 and frames["t"][0] < previous_t:
                    non_monotonic_ts += 1
                    previous_t = frames["t"][-1]
                non_monotonic_ts += numpy.count_nonzero(numpy.diff(frames["t"].astype("<i8")) < 0)
                size_mismatches = numpy.count_nonzero(
                    numpy.logical_or(frames["width"] != file.width, frames["height"] != file.height)
                )
        except RemainingBytesError as error:
            empty = False
            errors.add(f"{len(error.buffer)} extra bytes")
        if empty:
            errors.add("no data found")
        if non_monotonic_ts > 0:
            errors.add("{} non-monotonic timestamp{}".format(non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""))
        if size_mismatches > 0:
            errors.add("{} size mismatch{}".format(size_mismatches, "es" if size_mismatches > 1 else ""))
    return errors


def check_dvs_file(file: DvsFile) -> set[str]:
    errors = set()
    if file.width is None:
        errors.add("missing width")
    if file.height is None:
        errors.add("missing height")
    if len(errors) == 0:
        non_monotonic_ts = 0
        out_of_range_count = 0
        empty = True
        try:
            previous_t = 0
            for events in file.packets():
                empty = False
                if len(events) > 0 and events["t"][0] < previous_t:
                    non_monotonic_ts += 1
                    previous_t = events["t"][-1]
                non_monotonic_ts += numpy.count_nonzero(numpy.diff(events["t"].astype("<i8")) < 0)
                out_of_range_count += numpy.count_nonzero(
                    numpy.logical_or(events["x"] >= file.width, events["y"] >= file.height)
                )
        except RemainingBytesError as error:
            empty = False
            errors.add(f"{len(error.buffer)} extra bytes")
        if empty:
            errors.add("no data found")
        if out_of_range_count > 0:
            errors.add("{} event{} out of range".format(out_of_range_count, "s" if out_of_range_count > 1 else ""))
        if non_monotonic_ts > 0:
            errors.add("{} non-monotonic timestamp{}".format(non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""))
    return errors


def check_imu_file(file: ImuFile) -> set[str]:
    errors = set()
    non_monotonic_ts = 0
    empty = True
    try:
        previous_t = 0
        for imus in file.packets():
            empty = False
            if len(imus) > 0 and imus["t"][0] < previous_t:
                non_monotonic_ts += 1
                previous_t = imus["t"][-1]
            non_monotonic_ts += numpy.count_nonzero(numpy.diff(imus["t"].astype("<i8")) < 0)
    except RemainingBytesError as error:
        errors.add(f"{len(error.buffer)} extra bytes")
    if empty:
        errors.add("no data found")
    if non_monotonic_ts > 0:
        errors.add("{} non-monotonic timestamp{}".format(non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""))
    return errors


def check_other_file(file: GenericFile) -> set[str]:
    collections.deque(file.chunks(), maxlen=0)
    return set()


def check_local_directory(
    directory: IndexedDirectory,
    delete_ds_store: bool,
    format_index: bool,
) -> None:
    for path in sorted(directory.path.iterdir()):
        if path.is_file():
            if path.name == "-index.json":
                continue
            if path.suffix == ".lz":
                if not path.stem in directory.files and not path.stem in directory.other_files:
                    print(
                        format_error(
                            'the file "{}" is not listed in "{}" (compressed files must be listed without their .lz extension)'.format(
                                path, directory.path / "-index.json"
                            )
                        )
                    )
            elif not path.name in directory.files and not path.name in directory.other_files:
                if delete_ds_store and path.name == ".DS_Store":
                    path.unlink()
                    print(f"🦘 deleted {path}")
                else:
                    print(
                        format_error('the file "{}" is not listed in "{}"'.format(path, directory.path / "-index.json"))
                    )
        elif path.is_dir():
            if path.name in directory.directories:
                check_local_directory(
                    directory=directory.directories[path.name],
                    delete_ds_store=delete_ds_store,
                    format_index=format_index,
                )
            else:
                print(
                    format_error(
                        'the directory "{}" is not listed in "{}"'.format(path, directory.path / "-index.json")
                    )
                )
    if format_index:
        with open(directory.path / "-index.json") as json_index_file:
            raw_json_data = json_index_file.read()
        json_index = json.loads(raw_json_data)
        json_index_schema.validate(json_index)
        json_index["directories"].sort(key=lambda path: path["name"])
        json_index["files"].sort(key=lambda path: path["name"])
        json_index["other_files"].sort(key=lambda path: path["name"])
        new_raw_json_data = f"{json.dumps(json_index, indent=4, sort_keys=True)}\n"
        if raw_json_data != new_raw_json_data:
            with open(directory.path / "-index.json", "w") as json_index_file:
                json_index_file.write(new_raw_json_data)
            print("🦘 formatted {}".format(directory.path / "-index.json"))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Download UNDR datasets listed in a configuration file")
    subparsers = parser.add_subparsers(dest="command")
    init_parser = subparsers.add_parser("init", help="Generate a default undr.toml file")
    init_parser.add_argument("--configuration", "-c", default="undr.toml", help="UNDR configuration file path")
    install_parser = subparsers.add_parser("install", help="Provision the datasets")
    configuration_and_timeout(install_parser)
    install_parser.add_argument(
        "--force", "-f", action="store_true", help="(Re-)download files even if they already exist locally"
    )
    install_parser.add_argument(
        "--workers-count", "-w", type=check_positive, default=32, help="Number of parallel processes"
    )
    bibtex_parser = subparsers.add_parser("bibtex", help="Generate a BibTeX referencing all the datasets")
    configuration_and_timeout(bibtex_parser)
    bibtex_parser.add_argument(
        "--output", "-o", default=None, help="Specify an output file (defaults to standard output)"
    )
    doctor_parser = subparsers.add_parser("doctor", help="Check the format of a dataset")
    configuration_and_timeout(doctor_parser)
    doctor_parser.add_argument(
        "--workers-count", "-w", type=check_positive, default=32, help="Number of parallel processes"
    )
    check_local_directory_parser = subparsers.add_parser(
        "check-local-directory", help="Check the format of a local directory"
    )
    check_local_directory_parser.add_argument("path", help="Path to the local directory")
    check_local_directory_parser.add_argument(
        "--delete-ds-store", "-d", action="store_true", help="Delete macOS's .DS_Store files"
    )
    check_local_directory_parser.add_argument(
        "--format-index", "-i", action="store_true", help="Sort -index.json files and format them"
    )
    check_local_directory_parser.add_argument(
        "--workers-count", "-w", type=check_positive, default=32, help="Number of parallel processes"
    )
    args = parser.parse_args()

    if args.command == "init":
        target = pathlib.Path(args.configuration)
        if target.is_file():
            print(f"{target} already exists")
        else:
            shutil.copyfile(dirname / "undr_default.toml", target)

    if args.command == "install":
        configuration = Configuration(pathlib.Path(args.configuration), provision=False)
        if args.timeout is not None:
            for dataset in configuration.datasets.values():
                dataset.set_timeout(args.timeout, recursive=True)
        configuration.provision(force=args.force, quiet=False, workers_count=args.workers_count)

    if args.command == "bibtex":
        configuration = Configuration(pathlib.Path(args.configuration), provision=False)
        for dataset in configuration.datasets.values():
            if args.timeout is not None:
                dataset.set_timeout(args.timeout, recursive=True)
            dataset.mode = "remote"
        configuration.provision(force=False, quiet=True)
        bibtex = configuration.bibtex(pretty=True, timeout=args.timeout)
        if args.output is None:
            print(bibtex)
        else:
            with open(args.output, "wb") as bibtex_file:
                bibtex_file.write(bibtex.encode())

    if args.command == "doctor":
        configuration = Configuration(pathlib.Path(args.configuration), provision=False)
        for dataset in configuration.datasets.values():
            if args.timeout is not None:
                dataset.set_timeout(args.timeout, recursive=True)
            dataset.mode = "remote"
        configuration.provision(force=False, quiet=True, workers_count=args.workers_count)
        print(format_info("doctor"))
        for index, (name, dataset) in enumerate(configuration.datasets.items()):
            for _, results in dataset.recursive_map(
                prefix=f"{format_count(index, len(configuration.datasets))} {name}",
                handle_aps_file=check_aps_file,
                handle_dvs_file=check_dvs_file,
                handle_other_file=check_other_file,
                workers_count=args.workers_count,
            ):
                for file, errors in list(results):
                    for error in errors:
                        print(f'{format_error(error)} in "{file.path}"')

    if args.command == "check-local-directory":
        print(format_info("check local directory"))
        directory = IndexedDirectory(
            path=pathlib.Path(args.path),
            own_doi=None,
            server=server_factory(url=f"{pathlib.Path(args.path).as_posix()}/", timeout=10.0, type="local"),
            parent=None,
            metadata={},
            provision=True,
        )
        try:
            directory.download(force=False, prefix=None, workers_count=1)
        except FileNotFoundError as error:
            print(
                format_error(
                    '"{}" not found (listed in "{}")'.format(error, pathlib.Path(str(error)).parent / "-index.json")
                )
            )
        check_local_directory(directory, delete_ds_store=args.delete_ds_store, format_index=args.format_index)
        for _, results in directory.recursive_map(
            prefix=f"{directory.path}",
            handle_aps_file=check_aps_file,
            handle_dvs_file=check_dvs_file,
            handle_imu_file=check_imu_file,
            handle_other_file=check_other_file,
            workers_count=args.workers_count,
        ):
            for file, errors in list(results):
                for error in errors:
                    print(f'{format_error(error)} in "{file.path}"')



import os
from . import certificates

certificates_bundle = certificates.bundle()
if certificates_bundle is not None:
    os.environ["REQUESTS_CA_BUNDLE"] = certificates_bundle
import requests


def from_doi(doi: str, pretty: bool, timeout: float) -> str:
    response = requests.get(
        f"https://dx.doi.org/{doi}", timeout=timeout, headers={"Accept": "application/x-bibtex; charset=utf-8"}
    )
    response.raise_for_status()
    if pretty:
        bibtex = ""
        new_line = True
        depth = 0
        for character in response.text:
            if new_line:
                if not character.isspace():
                    new_line = False
                    bibtex += " " * ((depth - 1 if character == "}" else depth) * 4)
            if character == "{":
                depth += 1
                bibtex += "{"
            elif character == "}":
                depth -= 1
                bibtex += "}"
            elif character == "\n":
                new_line = True
                bibtex += "\n"
            elif character.isspace():
                if not new_line:
                    bibtex += character
            else:
                bibtex += character
        if not bibtex.endswith("\n"):
            bibtex += "\n"
        return bibtex
    return response.text



import atexit
import pathlib
import shutil
import sys
import tempfile
import typing


temporary_directory: typing.Optional[pathlib.Path] = None


def _cleanup() -> None:
    global temporary_directory
    if temporary_directory is not None:
        shutil.rmtree(temporary_directory)
        temporary_directory = None


atexit.register(_cleanup)


def bundle() -> str:
    global temporary_directory
    if temporary_directory is not None:
        return str(temporary_directory / "ca-bundle.crt")
    if sys.platform == "win32":
        return None
    if sys.platform == "darwin":
        # from https://github.com/sheagcraig/MacSesh/blob/master/macsesh/keychain.py
        from Security import (
            errSecSuccess,
            kSecClass,
            kSecReturnRef,
            kSecMatchLimit,
            kSecMatchLimitAll,
            SecItemCopyMatching,
            kSecClassCertificate,
            kSecMatchTrustedOnly,
            SecItemExport,
            kSecFormatUnknown,
            SecTrustCopyAnchorCertificates,
        )
        from urllib3.contrib._securetransport.bindings import CoreFoundation, Security

        certificates = []
        result_code, result = SecItemCopyMatching(
            {
                kSecClass: kSecClassCertificate,
                kSecReturnRef: True,
                kSecMatchLimit: kSecMatchLimitAll,
                kSecMatchTrustedOnly: True,
            },
            None,
        )
        if result_code == errSecSuccess:
            certificates.extend(result)
        result_code, result = SecTrustCopyAnchorCertificates(None)
        if result_code == errSecSuccess:
            certificates.extend(result)
        if len(certificates) == 0:
            return None
        return_code, pem_data = SecItemExport(certificates, kSecFormatUnknown, 0, None, None)
        if return_code != errSecSuccess:
            return None
        temporary_directory = pathlib.Path(tempfile.mkdtemp())
        with open(temporary_directory / "ca-bundle.crt", "wb") as certificates_bundle:
            certificates_bundle.write(pem_data)
        return str(temporary_directory / "ca-bundle.crt")

    # from https://golang.org/src/crypto/x509/root_linux.go
    for path in (
        "/etc/ssl/certs/ca-certificates.crt",  # Debian/Ubuntu/Gentoo etc.
        "/etc/pki/tls/certs/ca-bundle.crt",  # Fedora/RHEL 6
        "/etc/ssl/ca-bundle.pem",  # OpenSUSE
        "/etc/pki/tls/cacert.pem",  # OpenELEC
        "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",  # CentOS/RHEL 7
        "/etc/ssl/cert.pem",  # Alpine Linux
    ):
        if pathlib.Path(path).is_file():
            return str(path)
    certificates_paths = []
    for path in (
        "/etc/ssl/certs",  # SLES10/SLES11
        "/etc/pki/tls/certs",  # Fedora/RHEL
        "/system/etc/security/cacerts",  # Android
    ):
        if path.is_dir():
            for subpath in path.iterdir():
                if subpath.is_file():
                    certificates_paths.append(subpath)
            if len(certificates_path) > 0:
                temporary_directory = pathlib.Path(tempfile.mkdtemp())
                with open(temporary_directory / "ca-bundle.crt", "wb") as certificates_bundle:
                    for path in certificates_path:
                        with open(path, "rb") as certificate:
                            certificates_bundle.write(certificate.read())
                return str(temporary_directory / "ca-bundle.crt")
    return None



import collections
import copy
import dataclasses
import threading
import time
import typing
import sys
from . import utilities


@dataclasses.dataclass
class StatusUpdate:
    todo_file_count_delta: int = 0
    done_file_count_delta: int = 0
    todo_size_delta: int = 0
    done_size_delta: int = 0


@dataclasses.dataclass
class Status:
    todo_file_count: int = 0
    done_file_count: int = 0
    todo_size: int = 0
    done_size: int = 0

    def apply_update(self, status_update: StatusUpdate) -> None:
        self.todo_file_count = max(0, self.todo_file_count + status_update.todo_file_count_delta)
        self.done_file_count = max(0, self.done_file_count + status_update.done_file_count_delta)
        self.todo_size = max(0, self.todo_size + status_update.todo_size_delta)
        self.done_size = max(0, self.done_size + status_update.done_size_delta)


class Manager:
    def __init__(self, status: Status, output_interval: float, speed_samples_count: int):
        self._status = copy.copy(status)
        self._output_interval = output_interval
        self._speed_samples_count = speed_samples_count
        self._running = True
        self._speed_samples = collections.deque()
        self._begin = time.monotonic()
        self._begin_done_size = self._status.done_size
        self.queue = collections.deque()

        def target() -> None:
            previous_dispatch = time.monotonic()
            previous_done_size = self._status.done_size
            self.output(status=self._status, last_call=False, speed=0)
            while self._running:
                now = time.monotonic()
                if now > previous_dispatch + self._output_interval:
                    if len(self._speed_samples) >= self._speed_samples_count:
                        self._speed_samples.popleft()
                    self._speed_samples.append(
                        (self._status.done_size - previous_done_size) / (now - previous_dispatch)
                    )
                    previous_dispatch = now
                    previous_done_size = self._status.done_size
                    self.output(
                        status=self._status, last_call=False, speed=sum(self._speed_samples) / len(self._speed_samples)
                    )
                try:
                    message = self.queue.popleft()
                    if isinstance(message, Status):
                        self._status = message
                    elif isinstance(message, StatusUpdate):
                        self._status.apply_update(status_update=message)
                    else:
                        raise Exception(f"unsupported message type {message}")
                except IndexError:
                    time.sleep(output_interval)
                    pass
            while True:
                try:
                    message = self.queue.popleft()
                    if isinstance(message, Status):
                        self._status = message
                    elif isinstance(message, StatusUpdate):
                        self._status.apply_update(status_update=message)
                    else:
                        raise Exception(f"unsupported message type {message}")
                except IndexError:
                    break
            self.output(
                status=self._status,
                last_call=True,
                speed=(self._status.done_size - self._begin_done_size) / (time.monotonic() - self._begin),
            )

        self._worker = threading.Thread(target=target)
        self._worker.daemon = True
        self._worker.start()

    def output(self, status: Status, last_call: bool, speed: float) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        self._running = False
        self._worker.join()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class Printer(Manager):
    def __init__(
        self,
        prefix: str,
        status: Status,
        output_interval: float = 0.1,
        speed_samples_count: int = 30,
        file_object: typing.IO = None,
    ):
        self.prefix = prefix
        self.file_object = sys.stdout if file_object is None else file_object
        self.previous_message_length = 0
        super().__init__(status=status, output_interval=output_interval, speed_samples_count=speed_samples_count)

    def output(self, status: Status, last_call: bool, speed: float):
        total_size = status.todo_size + status.done_size
        total_count = status.todo_file_count + status.done_file_count
        if total_count == 0:
            message = self.prefix
        else:
            message = "{} - {} / {} file{}, {:.2f} % ({} / {}){}".format(
                self.prefix,
                status.done_file_count,
                total_count,
                "" if total_count == 1 else "s",
                0 if total_size == 0 else status.done_size / total_size * 100,
                utilities.size_to_string(status.done_size),
                utilities.size_to_string(total_size),
                f", {utilities.speed_to_string(speed)}" if speed > 0 else "",
            )
        self.file_object.write(
            "\r{}\r{}{}".format(
                " " * self.previous_message_length,
                message,
                "\n" if last_call else "",
            )
        )
        self.file_object.flush()
        self.previous_message_length = len(message)



import collections
import dataclasses
import datetime
import itertools
import json
import jsonschema_rs
import lzip
import lzip_extension
import math
import numpy
import os
import pathlib
import threading
import toml
import typing
import sys
from . import bibtex
from . import progress
from . import raw
from . import remote

dirname = pathlib.Path(__file__).resolve().parent

with open(dirname / "undr_schema.json") as undr_schema_file:
    undr_schema = jsonschema_rs.JSONSchema(json.load(undr_schema_file))

with open(dirname / "-index_schema.json") as index_schema_file:
    json_index_schema = jsonschema_rs.JSONSchema(json.load(index_schema_file))

ansi_colors_enabled = os.getenv("ANSI_COLORS_DISABLED") is None

RemainingBytesError = lzip.RemainingBytesError


def format_bold(message: str):
    if ansi_colors_enabled:
        return f"\033[1m{message}\033[0m"
    return message


def format_dim(message: str):
    if ansi_colors_enabled:
        return f"\033[2m{message}\033[0m"
    return message


def format_info(message: str):
    return f"🦘 {format_bold(message)}"


def format_error(message: str):
    return f"❌ {message}"


def format_count(index, total):
    total_as_string = str(total)
    return format_dim(f"({index + 1:>{len(total_as_string)}d} / {total_as_string})")


def server_factory(type: typing.Optional[str], **kwargs) -> remote.Server:
    if type is not None:
        if type == "apache":
            return remote.ApacheServer(**kwargs)
        if type == "local":
            return remote.LocalServer(**kwargs)
        if type == "nginx":
            return remote.ApacheServer(**kwargs)
        raise RuntimeError(f"unsupported server type {type}")
    return remote.Server(**kwargs)


class SerializableTomlDecoder(toml.TomlDecoder):
    def get_empty_inline_table(self):
        return self.get_empty_table()


@dataclasses.dataclass
class Path:
    path: pathlib.Path
    own_doi: typing.Optional[str]
    server: remote.Server
    parent: typing.Optional["IndexedDirectory"] = dataclasses.field(repr=False)
    metadata: dict[str, typing.Any] = dataclasses.field(repr=False)


@dataclasses.dataclass
class GenericFile(Path):
    remote_path: pathlib.PurePosixPath
    progress_queue: typing.Optional[collections.deque] = dataclasses.field(default=None, repr=False)

    def lzip_path(self) -> pathlib.Path:
        return pathlib.Path(str(self.path) + ".lz")

    def decompress_path(self) -> pathlib.Path:
        return pathlib.Path(str(self.path) + ".decompress")

    def as_resource(self) -> remote.Resource:
        return remote.Resource(local_path=self.path, remote_path=self.remote_path)

    def download(self, force: bool) -> None:
        self.server.download(resource=self.as_resource(), force=force, workers_count=1)

    def decompressible(self, force: bool) -> bool:
        return (not self.path.is_file() or force) and self.lzip_path().is_file()

    def decompress(self, force: bool) -> None:
        if self.decompressible(force):
            self.decompress_path().unlink(missing_ok=True)
            with open(self.decompress_path(), "wb") as decompress_file:
                decoder = lzip_extension.Decoder(1)
                with open(self.lzip_path(), "rb") as lzip_file:
                    while True:
                        encoded_bytes = lzip_file.read(65536)
                        if len(encoded_bytes) == 0:
                            break
                        if self.progress_queue is not None:
                            self.progress_queue.append(
                                progress.StatusUpdate(
                                    todo_size_delta=-len(encoded_bytes),
                                    done_size_delta=len(encoded_bytes),
                                )
                            )
                        decompress_file.write(decoder.decompress(encoded_bytes))
                decompress_file.write(decoder.finish()[0])
            self.decompress_path().rename(self.path)
            self.lzip_path().unlink()
            if self.progress_queue is not None:
                self.progress_queue.append(progress.StatusUpdate(todo_file_count_delta=-1, done_file_count_delta=1))

    def _chunks(self, word_size: int) -> typing.Iterator[bytes]:
        assert word_size > 0
        if self.path.is_file():
            chunk_size = math.ceil(65536 / word_size) * word_size
            with open(self.path, "rb") as file_object:
                while True:
                    chunk = file_object.read(chunk_size)
                    if len(chunk) == 0:
                        break
                    if len(chunk) % word_size != 0:
                        raise RemainingBytesError(word_size=word_size, buffer=chunk)
                    yield chunk
                    if self.progress_queue is not None:
                        self.progress_queue.append(
                            progress.StatusUpdate(
                                todo_size_delta=-len(chunk),
                                done_size_delta=len(chunk),
                            )
                        )
        elif self.lzip_path().is_file():
            decoder = lzip_extension.Decoder(word_size)
            with open(self.lzip_path(), "rb") as lzip_file:
                while True:
                    encoded_bytes = lzip_file.read(65536)
                    if len(encoded_bytes) == 0:
                        break
                    yield decoder.decompress(encoded_bytes)
                    if self.progress_queue is not None:
                        self.progress_queue.append(
                            progress.StatusUpdate(
                                todo_size_delta=-len(encoded_bytes),
                                done_size_delta=len(encoded_bytes),
                            )
                        )
                decoded_bytes, remaining_bytes = decoder.finish()
                if len(decoded_bytes) > 0:
                    yield decoded_bytes
                if len(remaining_bytes) > 0:
                    raise RemainingBytesError(word_size, remaining_bytes)
        else:
            with self.server.session() as local_session:
                resource, estimated_size, _ = self.server.resource_pick(
                    session=local_session, resource=self.as_resource(), try_alternatives=True
                )
                actual_size, chunks = self.server.resource_size_and_chunks(session=local_session, resource=resource)
                if self.progress_queue is not None and actual_size != estimated_size:
                    self.progress_queue.append(progress.StatusUpdate(todo_size_delta=actual_size - estimated_size))
                if resource.remote_path.suffix == ".lz":
                    decoder = lzip_extension.Decoder(word_size)
                    for chunk in chunks(65536):
                        yield decoder.decompress(chunk)
                        if self.progress_queue is not None:
                            self.progress_queue.append(
                                progress.StatusUpdate(
                                    todo_size_delta=-len(chunk),
                                    done_size_delta=len(chunk),
                                )
                            )
                    decoded_bytes, remaining_bytes = decoder.finish()
                    if len(decoded_bytes) > 0:
                        yield decoded_bytes
                    if len(remaining_bytes) > 0:
                        raise RemainingBytesError(word_size, remaining_bytes)
                else:
                    for chunk in chunks(math.ceil(65536 / word_size) * word_size):
                        if len(chunk) % word_size != 0:
                            raise RemainingBytesError(word_size=word_size, buffer=chunk)
                        yield chunk
                        if self.progress_queue is not None:
                            self.progress_queue.append(
                                progress.StatusUpdate(
                                    todo_size_delta=-len(chunk),
                                    done_size_delta=len(chunk),
                                )
                            )
        self.progress_queue.append(progress.StatusUpdate(todo_file_count_delta=-1, done_file_count_delta=1))
        self.progress_queue = None

    def chunks(self) -> typing.Iterable[bytes]:
        yield from self._chunks(word_size=1)

    def content_monolithic(self) -> bytes:
        return b"".join(self.chunks())

    def size(self) -> int:
        if self.path.is_file():
            return self.path.stat().st_size
        if self.lzip_path().is_file():
            return self.lzip_path().stat().st_size
        with self.server.session() as local_session:
            return self.server.resource_pick(session=local_session, resource=self.as_resource(), try_alternatives=True)[
                1
            ]


@dataclasses.dataclass
class File(GenericFile):
    original_name: typing.Optional[str] = None
    sensor: typing.Optional[str] = None
    scene: typing.Optional[str] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    date: typing.Optional[datetime.datetime] = None

    def __post_init__(self):
        for field in ["original_name", "sensor", "scene", "width", "height", "date"]:
            if field in self.metadata:
                setattr(self, field, self.metadata[field])
                del self.metadata[field]


@dataclasses.dataclass
class ApsFile(File):
    def packets(self) -> typing.Iterable[numpy.array]:
        dtype = numpy.dtype(raw.aps_dtype(self.width, self.height))
        for chunk in self._chunks(word_size=dtype.itemsize):
            yield numpy.frombuffer(chunk, dtype=dtype)

    def frames_monolithic(self) -> numpy.array:
        return numpy.frombuffer(self.content_monolithic(), dtype=raw.aps_dtype(self.width, self.height))


@dataclasses.dataclass
class DvsFile(File):
    def packets(self) -> typing.Iterable[numpy.array]:
        for chunk in self._chunks(word_size=raw.dvs.dtype.itemsize):
            yield raw.dvs.frombuffer(chunk)

    def events_monolithic(self) -> numpy.array:
        return raw.dvs.frombuffer(self.content_monolithic())


@dataclasses.dataclass
class ImuFile(File):
    def packets(self) -> typing.Iterable[numpy.array]:
        for chunk in self._chunks(word_size=raw.imu.dtype.itemsize):
            yield raw.imu.frombuffer(chunk)

    def imu_monolithic(self) -> numpy.array:
        return raw.imu.frombuffer(self.content_monolithic())


def file_factory(type: str, **kwargs):
    if type == "aps":
        return ApsFile(**kwargs)
    if type == "dvs":
        return DvsFile(**kwargs)
    if type == "imu":
        return ImuFile(**kwargs)
    raise RuntimeError(f"unsupported file type {type}")


@dataclasses.dataclass
class Task:
    file: File
    handle_file: typing.Callable[[typing.Any], typing.Any]


@dataclasses.dataclass
class IndexedDirectory(Path):
    provision: dataclasses.InitVar[bool]
    directories: dict[str, "IndexedDirectory"] = dataclasses.field(default=None, init=False, repr=False)
    files: dict[str, File] = dataclasses.field(default=None, init=False, repr=False)
    other_files: dict[str, GenericFile] = dataclasses.field(default=None, init=False, repr=False)

    def __post_init__(self, provision):
        if provision:
            self.provision(prefix=None)

    def provision(self, prefix: typing.Optional[str]) -> None:
        if prefix is not None:
            sys.stdout.write(prefix)
            sys.stdout.flush()
        self.path.mkdir(exist_ok=True)
        if not (self.path / "-index.json").is_file():
            self.server.download(
                remote.Resource.from_string(
                    local_path=self.path / "-index.json",
                    remote_path="-index.json",
                ),
                force=False,
                try_alternatives=False,
            )
        with open(self.path / "-index.json") as json_index_file:
            json_index = json.load(json_index_file)
        json_index_schema.validate(json_index)
        """
        all_names = [
            path["name"]
            for path in itertools.chain(json_index["directories"], json_index["files"], json_index["other_files"])
        ]
        if len(all_names) != len(set(all_names)):
            unique_names = set()
            for name in all_names:
                if name in unique_names:
                    raise Exception('duplicated name "{}" in "{}"'.format(name, self.path / "-index.json"))
                unique_names.add(name)
        """
        if "doi" in json_index:
            self.own_doi = json_index["doi"]
            del json_index["doi"]
        self.directories = {
            directory["name"]: IndexedDirectory(
                path=self.path / directory["name"],
                own_doi=directory["doi"] if "doi" in directory else None,
                server=self.server.clone_with_url(self.server.join_url(directory["name"], trailing_slash=True)),
                parent=self,
                metadata={key: value for key, value in directory.items() if key != "name" and key != "doi"},
                provision=False,
            )
            for directory in json_index["directories"]
        }
        del json_index["directories"]
        indent = "" if prefix is None else " " * ((len(prefix) - len(prefix.lstrip(" "))) + 4)
        if prefix is not None:
            sys.stdout.write("\n")
            sys.stdout.flush()
        for index, (name, directory) in enumerate(self.directories.items()):
            directory.provision(
                prefix=None if prefix is None else f"{indent}{format_count(index, len(self.directories))} {name}"
            )
        self.files = {
            file["name"]: file_factory(
                type=file["type"],
                path=self.path / file["name"],
                own_doi=file["doi"] if "doi" in file else None,
                server=self.server,
                parent=self,
                metadata={
                    key: value for key, value in file.items() if key != "name" and key != "doi" and key != "type"
                },
                remote_path=pathlib.PurePosixPath(file["name"]),
            )
            for file in json_index["files"]
        }
        del json_index["files"]
        self.other_files = {
            other["name"]: GenericFile(
                path=self.path / other["name"],
                own_doi=other["doi"] if "doi" in other else None,
                server=self.server,
                parent=self,
                metadata={key: value for key, value in other.items() if key != "name" and key != "doi"},
                remote_path=pathlib.PurePosixPath(other["name"]),
            )
            for other in json_index["other_files"]
        }
        del json_index["other_files"]
        self.metadata = {**self.metadata, **json_index}

    def download(self, force: bool, prefix: typing.Optional[str], workers_count: int) -> None:
        assert workers_count > 0
        workload = self.server.workload(
            resources=itertools.chain(
                (file.as_resource() for file in self.files.values()),
                (other_file.as_resource() for other_file in self.other_files.values()),
            ),
            force=force,
            try_alternatives=True,
            workers_count=workers_count,
        )
        if prefix is None:
            printer = None
        else:
            printer = progress.Printer(prefix=prefix, status=workload.status)
            workload.progress_queue = printer.queue
        self.server.consume(workload=workload, workers_count=workers_count)
        if printer is not None:
            printer.close()
        indent = "" if prefix is None else " " * ((len(prefix) - len(prefix.lstrip(" "))) + 4)
        for index, (name, directory) in enumerate(self.directories.items()):
            directory.download(
                force=force,
                prefix=None if prefix is None else f"{indent}{format_count(index, len(self.directories))} {name}",
                workers_count=workers_count,
            )

    def decompress(self, force: bool, prefix: typing.Optional[str]) -> None:
        status = progress.Status()
        workload = []
        for index, file in enumerate(itertools.chain(self.files.values(), self.other_files.values())):
            if file.decompressible(force=force):
                status.apply_update(
                    progress.StatusUpdate(
                        todo_file_count_delta=1,
                        todo_size_delta=file.lzip_path().stat().st_size,
                    )
                )
                workload.append(file)
            else:
                status.apply_update(
                    progress.StatusUpdate(
                        done_file_count_delta=1,
                        done_size_delta=file.path.stat().st_size,
                    )
                )
        if prefix is None:
            printer = None
        else:
            printer = progress.Printer(prefix=prefix, status=status)
        for file in workload:
            if printer is not None:
                file.progress_queue = printer.queue
            file.decompress(force=force)
            if printer is not None:
                file.progress_queue = None
        if printer is not None:
            printer.close()
        indent = "" if prefix is None else " " * ((len(prefix) - len(prefix.lstrip(" "))) + 4)
        for index, (name, directory) in enumerate(self.directories.items()):
            directory.decompress(
                force=force,
                prefix=None if prefix is None else f"{indent}{format_count(index, len(self.directories))} {name}",
            )

    def clear_server_cache(self, recursive: bool) -> None:
        self.server.clear_cache()
        if recursive and self.directories is not None:
            for directory in self.directories.values():
                directory.clear_server_cache(recursive=recursive, collect=False)

    def set_timeout(self, timeout: float, recursive: bool) -> None:
        self.server.set_timeout(timeout)
        if recursive and self.directories is not None:
            for directory in self.directories.values():
                directory.set_timeout(timeout)

    def doi_to_paths(self) -> dict[str, list[Path]]:
        doi_to_paths = {}
        if self.own_doi is not None:
            if self.own_doi in doi_to_paths:
                doi_to_paths[self.own_doi].append(self)
            else:
                doi_to_paths[self.own_doi] = [self]
        for file in itertools.chain(self.files.values(), self.other_files.values()):
            if file.own_doi is not None:
                if file.own_doi in doi_to_paths:
                    doi_to_paths[file.own_doi].append(file)
                else:
                    doi_to_paths[file.own_doi] = [file]
        for directory in self.directories.values():
            for doi, paths in directory.doi_to_paths().items():
                if doi in doi_to_paths:
                    doi_to_paths[doi].extend(paths)
                else:
                    doi_to_paths[doi] = paths
        return doi_to_paths

    def recursive_map(
        self,
        prefix: typing.Optional[str],
        workers_count: int,
        handle_aps_file: typing.Optional[typing.Callable[[ApsFile], typing.Any]] = None,
        handle_dvs_file: typing.Optional[typing.Callable[[DvsFile], typing.Any]] = None,
        handle_imu_file: typing.Optional[typing.Callable[[ImuFile], typing.Any]] = None,
        handle_other_file: typing.Optional[typing.Callable[[GenericFile], typing.Any]] = None,
    ) -> typing.Iterable[tuple["IndexedDirectory", typing.Iterable[tuple[File, typing.Any]]]]:
        assert workers_count > 0
        tasks: collections.deque[Task] = collections.deque()
        status = progress.Status()
        for file in self.files.values():
            selected = False
            if isinstance(file, ApsFile):
                if handle_aps_file is not None:
                    tasks.append(Task(file=file, handle_file=handle_aps_file))
                    selected = True
            elif isinstance(file, DvsFile):
                if handle_dvs_file is not None:
                    tasks.append(Task(file=file, handle_file=handle_dvs_file))
                    selected = True
            elif isinstance(file, ImuFile):
                if handle_imu_file is not None:
                    tasks.append(Task(file=file, handle_file=handle_imu_file))
                    selected = True
            else:
                raise Exception(f"unsupported file type {file}")
            if selected:
                status.apply_update(
                    progress.StatusUpdate(
                        todo_file_count_delta=1,
                        todo_size_delta=file.size(),
                    )
                )
        if handle_other_file is not None:
            for file in self.other_files.values():
                tasks.append(Task(file=file, handle_file=handle_other_file))
                status.apply_update(
                    progress.StatusUpdate(
                        todo_file_count_delta=1,
                        todo_size_delta=file.size(),
                    )
                )
        if prefix is None:
            printer = None
        else:
            printer = progress.Printer(prefix=prefix, status=status)
            for task in tasks:
                task.file.progress_queue = printer.queue
        if min(workers_count, len(tasks)) < 2:

            def worker_target():
                try:
                    while True:
                        task = tasks.popleft()
                        yield (task.file, task.handle_file(task.file))
                except IndexError:
                    pass
                if printer is not None:
                    printer.close()

            yield (self, worker_target())
        else:
            results: collections.deque[typing.Any] = collections.deque()
            result_available = threading.Condition()
            expected_results = len(tasks)

            def consume_results():
                with result_available:
                    try:
                        index = 0
                        while index < expected_results:
                            try:
                                yield results.popleft()
                                index += 1
                            except IndexError:
                                result_available.wait(0.1)
                    except IndexError:
                        pass
                if printer is not None:
                    printer.close()

            def worker_target():
                try:
                    while True:
                        task = tasks.popleft()
                        results.append((task.file, task.handle_file(task.file)))
                        with result_available:
                            result_available.notify()
                except IndexError:
                    pass

            workers = []
            for _ in range(0, min(workers_count, len(tasks))):
                worker = threading.Thread(target=worker_target)
                worker.daemon = True
                worker.start()
                workers.append(worker)
            yield (self, consume_results())
            for worker in workers:
                worker.join()
        del status
        del worker_target
        del printer
        del tasks
        self.clear_server_cache(recursive=False)
        indent = "" if prefix is None else " " * ((len(prefix) - len(prefix.lstrip(" "))) + 4)
        for index, (name, directory) in enumerate(self.directories.items()):
            yield from directory.recursive_map(
                prefix=None if prefix is None else f"{indent}{format_count(index, len(self.directories))} {name}",
                workers_count=workers_count,
                handle_aps_file=handle_aps_file,
                handle_dvs_file=handle_dvs_file,
                handle_imu_file=handle_imu_file,
                handle_other_file=handle_other_file,
            )


@dataclasses.dataclass
class Dataset(IndexedDirectory):
    mode: str = "remote"


@dataclasses.dataclass
class Configuration:
    directory: pathlib.Path = dataclasses.field(init=False)
    datasets: dict[str, IndexedDirectory] = dataclasses.field(init=False, repr=False)
    path: typing.Union[str, os.PathLike] = "undr.toml"
    provision: dataclasses.InitVar[bool] = True

    def __post_init__(self, provision):
        self.path = pathlib.Path(self.path).resolve()
        with open(self.path) as configuration_file:
            configuration = toml.load(configuration_file, decoder=SerializableTomlDecoder())
        undr_schema.validate(configuration)
        directory = pathlib.Path(configuration["directory"])
        if directory.is_absolute():
            self.directory = directory
        else:
            self.directory = self.path.parent / directory
        self.datasets = {
            dataset["name"]: Dataset(
                path=self.directory / dataset["name"],
                own_doi=dataset["doi"] if "doi" in dataset else None,
                server=server_factory(
                    type=dataset["server_type"] if "server_type" in dataset else None,
                    url=dataset["url"],
                    timeout=dataset["timeout"] if "timeout" in dataset else 10.0,
                ),
                parent=None,
                metadata={
                    key: value
                    for key, value in dataset.items()
                    if key != "name" and key != "doi" and key != "url" and key != "mode" and key != "server_type"
                },
                provision=False,
                mode=dataset["mode"],
            )
            for dataset in configuration["datasets"]
        }
        if provision:
            self.provision(force=False, quiet=True, workers_count=32)

    def provision(self, force: bool, quiet: bool, workers_count: int) -> None:
        self.directory.mkdir(parents=True, exist_ok=True)
        if not quiet:
            print(format_info("provision"))
        for index, (name, dataset) in enumerate(self.datasets.items()):
            dataset.provision(prefix=None if quiet else f"{format_count(index, len(self.datasets))} {name}")
        if not quiet:
            print()
            print(format_info("download"))
        download_count = sum(1 for dataset in self.datasets.values() if dataset.mode != "remote")
        for index, (name, dataset) in enumerate(self.datasets.items()):
            if dataset.mode != "remote":
                dataset.download(
                    force=force,
                    prefix=None if quiet else f"{format_count(index, download_count)} {name}",
                    workers_count=workers_count,
                )
                dataset.clear_server_cache(recursive=True)
        if not quiet:
            print()
            print(format_info("decompress"))
        decompress_count = sum(1 for dataset in self.datasets.values() if dataset.mode == "local-decompressed")
        for name, dataset in self.datasets.items():
            if dataset.mode == "local-decompressed":
                dataset.decompress(
                    force=force, prefix=None if quiet else f"{format_count(index, decompress_count)} {name}"
                )

    def doi_to_paths(self) -> dict[str, list[Path]]:
        doi_to_paths = {}
        for dataset in self.datasets.values():
            for doi, paths in dataset.doi_to_paths().items():
                if doi in doi_to_paths:
                    doi_to_paths[doi].extend(paths)
                else:
                    doi_to_paths[doi] = paths
        return doi_to_paths

    def bibtex(self, pretty: bool, timeout: float) -> str:
        result = ""
        for doi, paths in self.doi_to_paths().items():
            if len(result) > 0:
                result += "\n"
            result += "% {}\n".format(", ".join(path.path.relative_to(self.directory).as_posix() for path in paths))
            result += bibtex.from_doi(doi=doi, pretty=pretty, timeout=timeout)
        return result



import setuptools
import shutil
import pathlib

dirname = pathlib.Path(__file__).resolve().parent

shutil.rmtree(dirname / "undr", ignore_errors=True)
shutil.copytree(dirname / "source", dirname / "undr")
shutil.copyfile(dirname / "specification" / "undr_schema.json", dirname / "undr" / "undr_schema.json")
shutil.copyfile(dirname / "specification" / "-index_schema.json", dirname / "undr" / "-index_schema.json")
shutil.copyfile(dirname / "undr_default.toml", dirname / "undr" / "undr_default.toml")

with open("README.md") as file:
    long_description = file.read()

setuptools.setup(
    name="undr",
    version="0.0.7",
    url="https://github.com/neuromorphicsystems/undr",
    author="Alexandre Marcireau",
    author_email="alexandre.marcireau@gmail.com",
    description="Download and process Neuromorphic datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "jsonschema-rs>=0.9",
        "lzip>=1.0",
        "numpy>=1.20",
        "requests>=2.25",
        "toml>=0.10",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=["undr"],
    package_data={"": ["*.json", "*.toml"]},
)



import math


def parse_size(size: str) -> int:
    if size[-1] == "K":
        return round(float(size[:-1]) * 1024)
    if size[-1] == "M":
        return round(float(size[:-1]) * (1024 ** 2))
    if size[-1] == "G":
        return round(float(size[:-1]) * (1024 ** 3))
    if size[-1] == "T":
        return round(float(size[:-1]) * (1024 ** 4))
    return round(float(size))


def duration_to_string(delta: int) -> str:
    if delta < 180000000:
        return f'{"{:.0f}".format(math.floor(delta / 1e6))} s'
    if delta < 10800000000:
        return f'{"{:.0f}".format(math.floor(delta / 60e6))} min'
    if delta < 259200000000:
        return f'{"{:.0f}".format(math.floor(delta / 3600e6))} h'
    return f'{"{:.0f}".format(math.floor(delta / 86400e6))} days'


def size_to_string(size: int) -> str:
    if size < 1000:
        return f'{"{:.0f}".format(size)} B'
    if size < 1000000:
        return f'{"{:.2f}".format(size / 1000)} kB'
    if size < 1000000000:
        return f'{"{:.2f}".format(size / 1000000)} MB'
    if size < 1000000000000:
        return f'{"{:.2f}".format(size / 1000000000)} GB'
    return f'{"{:.2f}".format(size / 1000000000000)} TB'


def speed_to_string(speed: int) -> str:
    if speed < 1000:
        return f'{"{:.0f}".format(speed)} B/s'
    if speed < 1000000:
        return f'{"{:.2f}".format(speed / 1000)} kB/s'
    if speed < 1000000000:
        return f'{"{:.2f}".format(speed / 1000000)} MB/s'
    if speed < 1000000000000:
        return f'{"{:.2f}".format(speed / 1000000000)} GB/s'
    return f'{"{:.2f}".format(speed / 1000000000000)} TB/s'
